/*
 * Copyright (C) 1996-2003 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
__RCSID("$LAAS: server.c,v 1.19 2003/03/05 16:45:25 matthieu Exp $");

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <rpc/types.h>
#include <rpc/xdr.h>


#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_STREAMS
#  include <sys/stream.h>
#  include <sys/stropts.h>
#else
#  include <sys/ioctl.h>
#endif

#include <tcl.h>

#include "global.h"
#include "server.h"
#include "GDHE_packets.h"
#include "GDHE_xdr_prot.h"
#include "watchDog.h"
#include "request.h" 
#include "temps.h"

#define GDHE_PORT 3000
#define MAX_CONNECTIONS 30

/* prototypes des fonctions locales */
static int  serversock(int port);

/* Connexion des clients */
typedef struct GDHE_CLIENT {
    int socket;
    FILE *stream;
    XDR *xdr;
    Tcl_Channel tclChannel;
} GDHE_CLIENT;

static FILE *log;
static XDR *xdr_log = NULL, *xdr_sweep;
static GDHE_CLIENT clients[MAX_CONNECTIONS];
static int nb_clients;
static char *logFileName;

/*----------------------------------------------------------------------*/

/**
 ** Initialisation service TCP/XDR
 **/
int
init_xdr(char *name)
{
    int serveur;
    int i;

    /*
     * Creation log
     */
    if (name != NULL) {
        logFileName = strdup(name);
        log = fopen(logFileName, "w");
        if (log == NULL) {
          perror(logFileName);
          exit(-1);
        }
        printf("GDHE: log to %s\n", logFileName);
        xdr_log = (XDR *)Tcl_Alloc(sizeof(XDR));
        xdrstdio_create(xdr_log, log, XDR_ENCODE);
    } else {
        log = fopen("/dev/null", "w");
        if (log == NULL) {
            perror("/dev/null");
            exit(-1);
        }
        printf("GDHE: no log file\n");
    }
    xdr_sweep = (XDR *)Tcl_Alloc(sizeof(XDR));
    xdrstdio_create(xdr_sweep, stdout, XDR_FREE);

    /* 
     * Creation Socket serveur
     */
    serveur = serversock(GDHE_PORT);
    if (serveur < 0) {
        perror("serversock");
        exit(-2);
    }
    /*
     * Init du tableau des clients 
     */
    for (i = 0; i < MAX_CONNECTIONS; i++) {
        clients[i].socket = -1;
        clients[i].xdr = NULL;
        clients[i].stream = NULL;
	clients[i].tclChannel = NULL;
    }

    return(serveur);
} /* init_xdr */

/*----------------------------------------------------------------------*/

/**
 ** ferme toutes les connexions XDR et le fichier de log 
 **/
void
end_xdr(void)
{
    int i;
    
    for (i = 0; i < MAX_CONNECTIONS; i++) {
        if (clients[i].xdr != NULL) {
            xdr_destroy(clients[i].xdr);
            fclose(clients[i].stream);
        }
    } /* for */
    if (xdr_log != NULL) {
        xdr_destroy(xdr_log);
    }
    xdr_destroy(xdr_sweep);
    fclose(log);
}
 
/*----------------------------------------------------------------------*/

/**
 ** Remet a zero le fichier de log
 **/
void 
reset_log(void)
{
    char oldname[MAXPATHLEN];
    
    if (xdr_log != NULL) {
        xdr_destroy(xdr_log);
        Tcl_Free((char *)xdr_log);
        fclose(log);
    }
    sprintf(oldname, "%s.bak", logFileName);
    rename(logFileName, oldname);

    if (logFileName != NULL) {
        log = fopen(logFileName, "w");
        if (log == NULL) {
          perror(logFileName);
          exit(-1);
        }
        xdr_log = (XDR *)Tcl_Alloc(sizeof(XDR));
        xdrstdio_create(xdr_log, log, XDR_ENCODE);
    } 
}
/*----------------------------------------------------------------------*/

/**
 ** handler de la connexion d'un nouveau client 
 **/

void
connect_client(ClientData clientData, int mask)
{
    struct sockaddr addr;
    int i, len = sizeof(struct sockaddr);
    int socket;
    FILE *stream;
    XDR *xdr;
    struct in_addr *peer_addr;
    struct hostent *ht;
#if 0
    Tcl_Obj *objv[4];
    int objc = 0;
#else
    static char cmd[2048];
#endif
    
    if (nb_clients >= MAX_CONNECTIONS) {
        fprintf(stderr, "GDHE: too many clients\n");
        return;
    }
    socket = accept(socket_serveur, &addr, &len);
    if (socket < 0) {
        perror("GDHE: accept");
        return;
    }
    peer_addr = &(((struct sockaddr_in *)&addr)->sin_addr);
    ht = gethostbyaddr((char *)peer_addr, sizeof(struct in_addr), AF_INET);
    
    /* XXX sous linux peer_addr est raz si gethostbyaddr choue ???? */
    peer_addr = &(((struct sockaddr_in *)&addr)->sin_addr);

    /* Verif de l'autorisation de connexion */
#if 0
    objv[objc++] = Tcl_NewStringObj("clientAuthorize", -1);
    objv[objc++] = Tcl_NewStringObj(inet_ntoa(*peer_addr), -1);
    if (ht != NULL) {
	objv[objc++] = Tcl_NewStringObj(ht->h_name, -1);
    }
    if (Tcl_EvalObj(tclInterp, Tcl_Concat(objc, objv)) != TCL_OK) {
	fprintf(stderr, "GDHE: connection from %s refused "
		"by clientAuthorize: %s\n", 
		Tcl_GetStringResult(tclInterp));
	close(socket);
	return;
    }
#else
    snprintf(cmd, sizeof(cmd), "clientAuthorize %s %s", 
	     inet_ntoa(*peer_addr),
	     (ht!=NULL)?ht->h_name:"\"\"");
    if (Tcl_Eval(tclInterp, cmd) != TCL_OK) {
	fprintf(stderr, "GDHE: connection from %s refused "
		"by clientAuthorize: %s\n", 
		(ht!=NULL)?ht->h_name:inet_ntoa(*peer_addr), 
		Tcl_GetStringResult(tclInterp));
	close(socket);
	return;
    }
#endif	
    stream = fdopen(socket, "r");
    if (stream == NULL) {
        perror("GDHE: fdopen");
        return;
    }
    setbuf(stream, NULL);

    xdr = (XDR *)Tcl_Alloc(sizeof(XDR));
    xdrstdio_create(xdr, stream, XDR_DECODE);

    /* Recheche d'une connexion libre */
    for (i = 0; i < MAX_CONNECTIONS; i++) {
        if (clients[i].socket < 0) {
            break;
        }
    } /* for */
    if (i == MAX_CONNECTIONS) {
        fprintf(stderr, "GDHE: connect_client: assertion failed"
                " (i< MAX_CONNECTIONS)");
    }
    /* Enregistre un handler pour l'arrive'e de donne'es 
       sur cette connexion */
    Tcl_CreateFileHandler(socket, TCL_READABLE | TCL_EXCEPTION,
			  traite_data, (ClientData)(long)i);

    /* Me'morise la connexion */
    clients[i].socket = socket;
    clients[i].stream = stream;
    clients[i].xdr = xdr;

    printf("GDHE: client %d connected from %s\n", i, 
           ht == NULL ? inet_ntoa(*peer_addr) : ht->h_name);

    nb_clients++;
    
} /* connect_client */

/*----------------------------------------------------------------------*/

/**
 ** traitement des donne'es 
 **/

void
traite_data(ClientData clientData, int mask)
{
    int client = (int)(long)clientData;
    XDR *xdr;
    static packet data;
    int n;
    static int in_traite_data = 0;
    int file;
    struct timeval current;

    if (in_traite_data == 1) {
        printf("GDHE: recursive call of traite_data()\n");
        return;
    } 
    in_traite_data = 1;

    file = clients[client].socket;
    xdr = clients[client].xdr;
    
    if (xdr == NULL) {
        printf("GDHE: traite_data called for a disconnected client: %d\n",
               client);
	Tcl_DeleteFileHandler(clients[client].socket);
        in_traite_data = 0;
        return;
    }

    /*
     * boucle de lecture et traitement de tous les paquets arrive's 
     */
    while (1) {
        /* Reception d'un paquet avec un watchdog a 10s */
        watchDogArm(10000, NULL, NULL);
        n = xdr_packet(xdr, &data);

        watchDogRelease();
        if (n == 0) {
	    Tcl_DeleteFileHandler(clients[client].socket);
            perror("GDHE: xdr_packet");
            fprintf(stderr, "      client %d disconnected\n", client);
            xdr_packet(xdr_sweep, &data);
            xdr_destroy(xdr);
            Tcl_Free((char *)xdr);
            fclose(clients[client].stream);
            close(clients[client].socket);
            clients[client].xdr = NULL;
            clients[client].stream = NULL;
            clients[client].socket = -1;
            nb_clients--;
            in_traite_data = 0;
            return;
        } 
	gettimeofday(&current, NULL);

        /* Sauvegarde dans le fichier de log */
        if (xdr_log != NULL) {
	    xdr_timeval(xdr_log, &current);
            xdr_packet(xdr_log, &data);
        }
        if (data.command == DISCONNECT) {
            fprintf(stderr, "GDHE: client %d disconnected\n", client);
	    Tcl_DeleteFileHandler(clients[client].socket);
            xdr_destroy(xdr);
            Tcl_Free((char *)xdr);
            fclose(clients[client].stream);
            close(clients[client].socket);
            clients[client].xdr = NULL;
            clients[client].stream = NULL;
            clients[client].socket = -1;
            nb_clients--;
        } else {
            request_exec(&data);
        }
        /* liberation du paquet */
        xdr_packet(xdr_sweep, &data);

        if (clients[client].socket >= 0) {
#ifdef HAVE_STREAMS
            if (ioctl(file, I_NREAD, &n) < 0) {
                perror("GDHE: ioctl");
                n = 0;
            }
#else
            if (ioctl(file, FIONREAD, &n) < 0) {
                perror("GDHE: ioctl");
                n = 0;
            }
#endif
            if (n == 0) {
		/* S'il n'y a plus rien a`, on rend la main */
                break;
            }
        } else {
	    /* De'connecte' */
            break;
        }
	if (need_redisplay(&current) && auto_redisplay) {
	    /* force un reaffichage */
	    Tcl_Eval((Tcl_Interp *)tclInterp, "redrawAllWindows true");
	}
    } /* while */

    
    in_traite_data = 0;
    if (auto_redisplay) {
	Tcl_Eval((Tcl_Interp *)tclInterp, "redrawAllWindows false");
    }

} /* traite_data */

/*----------------------------------------------------------------------*/

/*
** serversock()
**
** Creates an internet socket, binds it to an address, and prepares it for
** subsequent accept() calls by calling listen().
**
** Input: port number desired, or 0 for a random one
** Output: file descriptor of socket, or a negative error
*/
static int 
serversock(int port)
{
    int sock, x;
    struct      sockaddr_in server;
    int reuseSocket = 1;
    struct linger sockLinger;
    
    sock = socket(PF_INET, SOCK_STREAM, 6);
    if (sock < 0)
        return -errno;
    
   /*
     * Options du socket
     */
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                   (char *)&reuseSocket, sizeof(reuseSocket)) < 0) {
        perror("GDHE: setsockopt REUSEADDR");
        return(-1);
    }
 
    sockLinger.l_onoff = 0;
    sockLinger.l_linger = 0;
    if (setsockopt(sock, SOL_SOCKET, SO_LINGER,
                   (char *)&sockLinger, sizeof(sockLinger)) < 0) {
        perror("GDHE: setsockopt NO LINGER");
        return(-1);
    }
    memset(&server, 0, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons(port);
    
    x = bind(sock, (struct sockaddr *)&server, sizeof(server));
    if (x < 0) {
        perror("GDHE: bind");
        close(sock);
        return(-1);
    }
    
    listen(sock, 5);
    
    return sock;
}


